//**********************************************************************
// This source code is provided as an example.  No warranty is expressed
// or implied.  It is your responsibility to determine its suitability
// to your application, to modify it as necessary, and to validate
// its correctness.
//
// Licensed Materials - Property of IBM
// (c) Copyright IBM Corporation 2000. All Rights Reserved.
//
// US Government Users Restricted Rights - Use, duplication, or
// disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
//**********************************************************************
// Revision / History
// Defect     Date       Who      Description
// ---------- 01/03/2000 ldbader  Module created
//**********************************************************************

package com.ibm.etill.ldbcardcassette;


import  java.sql.Connection;
import  java.sql.PreparedStatement;
import  java.sql.ResultSet;
import  java.sql.SQLException;
import  java.util.Vector;

import  com.ibm.etill.framework.archive.ETillArchive;
import  com.ibm.etill.framework.cassette.APIResponse;
import  com.ibm.etill.framework.cassette.CassetteTransaction;
import  com.ibm.etill.framework.cassette.RefundRequest;
import  com.ibm.etill.framework.cassette.RefundReversalRequest;
import  com.ibm.etill.framework.log.ErrorLog;
import  com.ibm.etill.framework.log.ETillCassetteException;
import  com.ibm.etill.framework.log.Trace;
import  com.ibm.etill.framework.payapi.Batch;
import  com.ibm.etill.framework.payapi.Credit;
import  com.ibm.etill.framework.payapi.ETillAbortOperation;
import  com.ibm.etill.framework.payapi.Order;
import  com.ibm.etill.framework.payapi.ParameterTable;
import  com.ibm.etill.framework.supervisor.Supervisor;


/***********************************************************************
* The cassette credit object is responsible for performing all the
* cassette's credit requests (Refund, RefundReversal) and to manage
* the state of the framework credit object.
***********************************************************************/
public final class LdbCardCredit
   implements CassetteTransaction, LdbCardConstants {

   private static final String COPYRIGHT =
      "(c) Copyright IBM Corporation 2000 - All Rights Reserved.";

   //<gutter>
   //</gutter>

   //*******************************************************************
   // This is the cassette credit object's instance data.
   //******************************************************************/
   /********************************************************************
   * This cassette credit attribute contains the reference back to the
   * associated framework credit object.
   ********************************************************************/
   private Credit            objFrameworkCredit     = null;
   /********************************************************************
   * This cassette credit attribute contains the approval code.  The
   * approval code is supplied by the payment engine when a credit
   * transaction is successfully created and is used in subsequent
   * payment engine requests (like void) to reference this credit.
   ********************************************************************/
   private String            objApprovalCode        = null;
   /********************************************************************
   * This cassette credit attribute contains the decline reason. The
   * decline reason, originally small and therefore cryptic, was
   * designed to be displayed on a card swipe terminal at a point of
   * sale terminal.  In this cassette, it provides detailed information
   * about the cause of the failure.  (Since the text is provided by
   * the financial network, different for each network, and not
   * standardized in any way, it is impossible to translate these text
   * strings into return codes.)
   ********************************************************************/
   private String            objDeclineReason       = null;

   /********************************************************************
   * This constructor initializes a new cassette credit object.  A
   * reference to the associated framework's credit object is set.
   *
   * @param     argFrameworkCredit - The framework's credit object.
   * @param     argProtocolData - The protocol data parameter values
   *            provided by the merchant system.
   ********************************************************************/
   public LdbCardCredit(
      Credit          argFrameworkCredit,
      ParameterTable  argProtocolData
   ) {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.LdbCardCredit()"
         );
      }

      this.objFrameworkCredit  = argFrameworkCredit;

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.LdbCardCredit()"
         );
      }
   }

   /********************************************************************
   * This constructor initializes an existing cassette credit object.
   * A reference to the associated framework's credit object is set
   * and the remaining attributes are resurrected from the cassette's
   * credit database table.
   *
   * @param     argFrameworkCredit - The framework's credit object.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public LdbCardCredit(
      Credit          argFrameworkCredit
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.LdbCardCredit()"
         );
      }

      this.objFrameworkCredit  = argFrameworkCredit;
      this.resurrectCredit();

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.LdbCardCredit()"
         );
      }
   }

   /********************************************************************
   * This method is called by the framework when the merchant attempts
   * to Refund a credit that has previously been declined.  Usually
   * an Refund creates a new credit.  In this case, the creidt exists
   * and must be reset to its initial state so that the Refund can
   * be attempted again.
   ********************************************************************/
   public void resetTransaction() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.resetTransaction()"
         );
      }

      this.objApprovalCode  = null;
      this.objDeclineReason = null;


      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.resetTransaction()"
         );
      }
   }

   /********************************************************************
    * This method is used by the cassette object to resurrect an
    * existing cassette credit from the cassette credit database table.
    * <p>
    * This method is <b>not</b> part of the Archivable interface.  The
    * methods for the Archivable interface are passed a JDBC database
    * connection for the operation.  In this case, an ETillArchive
    * object is constructed to provide a JDBC databaes connection.
    *
    * @exception ETillAbortOperation - any failure results in this
    * exception.
    *******************************************************************/
   private void resurrectCredit() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.resurrectCredit()"
         );
      }

      // The SQL select statement to obtain the cassette credit data.
      String varSelect =
         " SELECT * FROM " +
         ETillArchive.geteTillDatabaseOwner() + "." + TBL_CREDIT +
         " WHERE " +
         CRE_MERCHANT + " = '" +
         this.objFrameworkCredit.getMerchantName() +
         "' AND " +
         CRE_ORDER    + " = '" +
         this.objFrameworkCredit.getOrderNumber() +
         "' AND " +
         CRE_CREDIT   + " = '" +
         this.objFrameworkCredit.getTransactionNumber() + "'";

      // Construct a JDBC connection.
      ETillArchive varDatabase = new ETillArchive();

      // Lock the JDBC connection.
      synchronized (ETillArchive.getArchiveLock()) {
         try {
            // Trace database read.
            if (Trace.isAnyoneTracing()) {
               Trace.traceDatabaseRead(
                  TRACE_ID,
                  "\n" + varSelect
               );
            }
            // Execute the query.
            ResultSet varResult = varDatabase.beginQuery(varSelect);

            if (varResult.next()) {
               // The required data was found.  Initialize this
               // object's instance data.
               this.setApproval(
                  ETillArchive.readString(varResult, CRE_APPROVAL)
               );
               this.setReason(
                  ETillArchive.readString(varResult, CRE_REASON)
               );
               // Assertion test:  If there is more than one row in the
               // result set, there is a serious defect and the
               // cassette cannot continue.  There should never be more
               // than one row for a given cassette credit object.
               // The error is logged and a cassette exception is thrown.
               if (varResult.next()) {
                  if (Trace.isAnyoneTracing()) {
                     Trace.traceErrorOccurred(
                        TRACE_ID,
                        "LdbCardPayment Found too many rows." +
                        "\n\tPRC = " + PRC_DATABASE_ERROR +
                        "\n\tSRC = " + SRC_CREDIT_SELECT_ROW_FAILURE
                     );
                  }
                  Vector varVector = new Vector();
                  varVector.addElement(
                     this.objFrameworkCredit.getMerchantName()
                  );
                  varVector.addElement(
                     this.objFrameworkCredit.getOrderNumber()
                  );
                  varVector.addElement(
                     this.objFrameworkCredit.getTransactionNumber()
                  );
                  ErrorLog.log(
                     RB_ID,
                     MSG_CREDIT_SELECT_ROW_FAILURE,
                     varVector
                  );
                  throw new ETillCassetteException(
                     ErrorLog.lookupMessage(
                        RB_ID,
                        MSG_CREDIT_SELECT_ROW_FAILURE,
                        varVector
                     )
                  );
               }
            } else {
               // The data was not found.  Log an error and abort.
               if (Trace.isAnyoneTracing()) {
                  Trace.traceErrorOccurred(
                     TRACE_ID,
                     "LdbCardCredit Credit resurrection failed." +
                     "\n\tPRC = " + PRC_DATABASE_ERROR +
                     "\n\tSRC = " + SRC_CREDIT_SELECT_CREDIT_MISSING
                  );
               }
               ErrorLog.logError(
                  RB_ID,
                  MSG_CREDIT_SELECT_CREDIT_MISSING,
                  null,
                  this.objFrameworkCredit.getMerchantName(),
                  this.objFrameworkCredit.getOrderNumber(),
                  this.objFrameworkCredit.getTransactionNumber()
               );
               throw new ETillAbortOperation(
                  PRC_DATABASE_ERROR,
                  SRC_CREDIT_SELECT_CREDIT_MISSING
               );
            }
         } catch (SQLException varException) {
            // The query failed.  Log the error and abort.
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardCredit SQLException caught." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_CREDIT_SELECT_SQL_FAILURE
               );
            }
            ErrorLog.logError(
               RB_ID,
               MSG_CREDIT_SELECT_SQL_FAILURE,
               varException,
               this.objFrameworkCredit.getMerchantName(),
               this.objFrameworkCredit.getOrderNumber(),
               this.objFrameworkCredit.getTransactionNumber()
            );
            throw new ETillAbortOperation(
               PRC_DATABASE_ERROR,
               SRC_CREDIT_SELECT_SQL_FAILURE
            );
         } finally {
            // Regardless of whether there was an exception or not,
            // clean up the ETillArchive object.
            try {
               varDatabase.endQuery();
            } catch (SQLException varException) {
               if (Trace.isAnyoneTracing()) {
                  Trace.traceErrorOccurred(
                     TRACE_ID,
                     "LdbCardPayment SQL Exception caught." +
                     "\n\tPRC = " + PRC_DATABASE_ERROR +
                     "\n\tSRC = " + SRC_CREDIT_SELECT_CLOSE_FAILURE
                  );
               }
               ErrorLog.logError(
                  RB_ID,
                  MSG_CREDIT_SELECT_CLOSE_FAILURE,
                  varException,
                  this.objFrameworkCredit.getMerchantName(),
                  this.objFrameworkCredit.getOrderNumber(),
                  this.objFrameworkCredit.getTransactionNumber()
               );
            }
         }
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.resurrectCredit()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework order (a commitable object)
   * when its commit() method discovers the cassette credit object on
   * the framework order's create list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void createRecord(
      Connection argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.createRecord()"
         );
      }

      try {
         // The SQL statement to insert a new cassette credit.
         String varInsert =
            " INSERT INTO " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_CREDIT + " (" +
            CRE_MERCHANT     + ", " +
            CRE_ORDER        + ", " +
            CRE_CREDIT       + ", " +
            CRE_APPROVAL     + ", " +
            CRE_REASON       +
            ") VALUES (?, ?, ?, ?, ?)";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varInsert + "\n" +
               CRE_MERCHANT  + " = " +
               this.objFrameworkCredit.getMerchantName() + "\n" +
               CRE_ORDER     + " = " +
               this.objFrameworkCredit.getOrderNumber() + "\n" +
               CRE_CREDIT    + " = " +
               this.objFrameworkCredit.getTransactionNumber() + "\n" +
               CRE_APPROVAL  + " = " + this.getApproval() + "\n" +
               CRE_REASON    + " = " + this.getReason() + "\n"
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varInsert);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(
            this.objFrameworkCredit.getMerchantName(),   varStatement, 1
         );
         ETillArchive.addString(
            this.objFrameworkCredit.getOrderNumber(),    varStatement, 2
         );
         ETillArchive.addString(
            this.objFrameworkCredit.getTransactionNumber(),
                                                         varStatement, 3
         );
         ETillArchive.addString(
            this.getApproval(),                          varStatement, 4
         );
         ETillArchive.addString(
            this.getReason(),                            varStatement, 5
         );
         // Execute the insert.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardCredit Created too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_CREDIT_CREATE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkCredit.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkCredit.getOrderNumber()
            );
            varVector.addElement(
               this.objFrameworkCredit.getTransactionNumber()
            );
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_CREDIT_CREATE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_CREDIT_CREATE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The insert failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCredit SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_CREDIT_CREATE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_CREDIT_CREATE_SQL_FAILURE,
            varException,
            this.objFrameworkCredit.getMerchantName(),
            this.objFrameworkCredit.getOrderNumber(),
            this.objFrameworkCredit.getTransactionNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_CREDIT_CREATE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.createRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework order (a commitable object)
   * when its commit() method discovers the cassette credit object on
   * the framework order's update list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void updateRecord(
      Connection argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.updateRecord()"
         );
      }

      try {
         // The SQL statement to update an existing cassette credit.
         String varUpdate =
            " UPDATE " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_CREDIT + " SET " +
            CRE_APPROVAL     + "= ?, " +
            CRE_REASON       + "= ? "  +
            " WHERE " +
            CRE_MERCHANT + " = ? AND " +
            CRE_ORDER    + " = ? AND " +
            CRE_CREDIT   + " = ?";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varUpdate + "\n" +
               CRE_APPROVAL  + " = " + this.getApproval() + "\n" +
               CRE_REASON    + " = " + this.getReason() + "\n" +
               CRE_MERCHANT  + " = " +
               this.objFrameworkCredit.getMerchantName() + "\n" +
               CRE_ORDER     + " = " +
               this.objFrameworkCredit.getOrderNumber() + "\n" +
               CRE_CREDIT    + " = " +
               this.objFrameworkCredit.getTransactionNumber() + "\n"
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varUpdate);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(
            this.getApproval(),                          varStatement, 1
         );
         ETillArchive.addString(
            this.getReason(),                            varStatement, 2
         );
         ETillArchive.addString(
            this.objFrameworkCredit.getMerchantName(),   varStatement, 3
         );
         ETillArchive.addString(
            this.objFrameworkCredit.getOrderNumber(),    varStatement, 4
         );
         ETillArchive.addString(
            this.objFrameworkCredit.getTransactionNumber(),
                                                         varStatement, 5
         );
         // Execute the update.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardCredit Modified too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_CREDIT_UPDATE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkCredit.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkCredit.getOrderNumber()
            );
            varVector.addElement(
               this.objFrameworkCredit.getTransactionNumber()
            );
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_CREDIT_UPDATE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_CREDIT_UPDATE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The update failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCredit SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_CREDIT_UPDATE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_CREDIT_UPDATE_SQL_FAILURE,
            varException,
            this.objFrameworkCredit.getMerchantName(),
            this.objFrameworkCredit.getOrderNumber(),
            this.objFrameworkCredit.getTransactionNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_CREDIT_UPDATE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.updateRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the framework order (a commitable object)
   * when its commit() method discovers the cassette credit object on
   * the framework order's delete list.
   *
   * @param     argConnection - The JDBC database connection to be used
   * for this operation.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void deleteRecord(
      Connection argConnection
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.deleteRecord()"
         );
      }

      try {
         // The SQL statement to delete an existing cassette credit.
         String varDelete =
            " DELETE FROM " +
            ETillArchive.geteTillDatabaseOwner() + "." + TBL_CREDIT +
            " WHERE " +
            CRE_MERCHANT + " = ? AND " +
            CRE_ORDER    + " = ? AND " +
            CRE_CREDIT   + " = ?";
         if (Trace.isAnyoneTracing()) {
            Trace.traceDatabaseWrite(
               TRACE_ID,
               "\n" + varDelete + "\n" +
               CRE_MERCHANT + " = " +
               this.objFrameworkCredit.getMerchantName() + "\n" +
               CRE_ORDER    + " = " +
               this.objFrameworkCredit.getOrderNumber() + "\n" +
               CRE_CREDIT   + " = " +
               this.objFrameworkCredit.getTransactionNumber() + "\n"
            );
         }
         // Ask the connection to build a prepared statement using
         // the SQL statement defined above.
         PreparedStatement varStatement =
            argConnection.prepareStatement(varDelete);
         // Clear all the variable parameters from the statement.
         varStatement.clearParameters();
         // Tell the statement to set the variable values.
         ETillArchive.addString(
            this.objFrameworkCredit.getMerchantName(),   varStatement, 1
         );
         ETillArchive.addString(
            this.objFrameworkCredit.getOrderNumber(),    varStatement, 2
         );
         ETillArchive.addString(
            this.objFrameworkCredit.getTransactionNumber(),
                                                         varStatement, 3
         );
         // Execute the delete.
         int varRowsAffected = varStatement.executeUpdate();
         // Close the statement.
         varStatement.close();
         // Assertion test:  If the number of affected rows is not 1
         // there is a serious defect and the cassette cannot continue.
         // The error is logged and a cassette exception is thrown.
         if (varRowsAffected != 1) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceErrorOccurred(
                  TRACE_ID,
                  "LdbCardCredit Deleted too many rows." +
                  "\n\tPRC = " + PRC_DATABASE_ERROR +
                  "\n\tSRC = " + SRC_CREDIT_DELETE_ROW_FAILURE
               );
            }
            Vector varVector = new Vector();
            varVector.addElement(
               this.objFrameworkCredit.getMerchantName()
            );
            varVector.addElement(
               this.objFrameworkCredit.getOrderNumber()
            );
            varVector.addElement(
               this.objFrameworkCredit.getTransactionNumber()
            );
            varVector.addElement(Integer.toString(varRowsAffected));
            ErrorLog.log(
               RB_ID,
               MSG_CREDIT_DELETE_ROW_FAILURE,
               varVector
            );
            throw new ETillCassetteException(
               ErrorLog.lookupMessage(
                  RB_ID,
                  MSG_CREDIT_DELETE_ROW_FAILURE,
                  varVector
               )
            );
         }
      }  catch (SQLException varException) {
         // The delete failed.  Log the error and abort.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCredit SQLException caught." +
               "\n\tPRC = " + PRC_DATABASE_ERROR +
               "\n\tSRC = " + SRC_CREDIT_DELETE_SQL_FAILURE
            );
         }
         ErrorLog.logError(
            RB_ID,
            MSG_CREDIT_DELETE_SQL_FAILURE,
            varException,
            this.objFrameworkCredit.getMerchantName(),
            this.objFrameworkCredit.getOrderNumber(),
            this.objFrameworkCredit.getTransactionNumber()
         );
         throw new ETillAbortOperation(
            PRC_DATABASE_ERROR,
            SRC_CREDIT_DELETE_SQL_FAILURE
         );
      }

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.deleteRecord()"
         );
      }
   }

   /********************************************************************
   * This method is used by the order object when it is asked to service
   * a Refund request.
   * <p>
   * For this cassette, the account object is used to authorize the
   * credit.  The framework credit object's state is updated and
   * return codes are placed in the provided response object.
   *
   * @param     argRequest - An Credit order request.
   * @param     argResponse - The response object to hold the result.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void refund(
      RefundRequest          argRequest,
      APIResponse            argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.refund()"
         );
      }

      // Abort Refund if the Credit is not reset, declined, or void
      if (
          this.objFrameworkCredit.getCurrentState() != CREDIT_RESET &&
          this.objFrameworkCredit.getCurrentState() != CREDIT_DECLINED &&
          this.objFrameworkCredit.getCurrentState() != CREDIT_VOID
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCredit Refund was requested but the " +
               " credit is not in CREDIT_RESET, CREDIT_DECLINED, " +
               " or CREDIT_VOID state." +
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_CREDIT +
               "\n\tState  = " + this.objFrameworkCredit.getCurrentState()
            );
         }
         throw new ETillAbortOperation(
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_CREDIT
         );
      }
      // Obtain the framework's order object from the request.
      Order varFrameworkOrder = argRequest.getOrder();
      // Obtain the cassette order object from the framework order
      // object.
      LdbCardOrder varCassetteOrder =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Obtain the account object that is used to process
      // transactions for this order.
      LdbCardAccount varAccount = varCassetteOrder.getAccount();
      // Ask the account to authorize this credit.  If it returns
      // True, the authorization was approved.  Otherwise, the
      // authorization was declined.  The account will set this
      // object's approval code and decline reason before it returns.
      if (varAccount.captureCredit(this)) {
         // The capture was approved.  Update the framework object,
         // add the credit to the batch, and set return codes.
         // It is the cassette's responsibility to set the credit
         // reference number to a value that the merchant can use to
         // identify the credit when reading a report sent by the
         // payment processor.
         this.objFrameworkCredit.setReferenceNumber(this.objApprovalCode);
         // The authorizaton was approved.  Update the state
         // information in the framework credit object and set
         // the correct return codes into the response.
         this.objFrameworkCredit.setCurrentState(this.CREDIT_REFUNDED);
         // Since the capture was approved, add this credit to
         // the current batch and add the framework batch to the
         // commit point.
         varAccount.selectBatchAndAddCredit(this.objFrameworkCredit);
         argResponse.setReturnCodes(
            PRC_OPERATION_SUCCESS,
            RC_NONE
         );
      } else {
         // The authorizaton was declined.  Update the state
         // information in the framework credit object and set
         // the correct return codes into the response.
         this.objFrameworkCredit.setCurrentState(CREDIT_DECLINED);
         // RC_CASSETTE_DECLINED
         // RC_CASSETTE_DECLINED_EXPIRY
         // RC_CASSETTE_DECLINED_INSTRUMENT
         argResponse.setReturnCodes(
            PRC_FINANCIAL_FAILURE,
            RC_CASSETTE_DECLINED
         );
         // Trace the error.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCredit Credit declined. " +
               "\n\tPRC = " + PRC_FINANCIAL_FAILURE +
               "\n\tSRC = " + RC_CASSETTE_DECLINED +
               "\n\tReason = " + this.getReason()
            );
         }
      }
      // There are two cases.  Either this is a new credit and it
      // must be added to the create list or it is a refund for
      // a credit that was previously declined and it must be added
      // to the update list.  The framework credit object contains
      // a boolean variable that can be tested to determine which
      // case this is.
      if (this.objFrameworkCredit.isNewTransaction())  {
         // Add the framework credit object to the framework
         // order's create list.  When the framework credit's
         // createRecord() method is called, it will call the
         // cassette credit's createRecord() method.
         Supervisor.getThreadCommitPoint().
            addToCreateList(this.objFrameworkCredit);
      } else {
         // Add the framework credit object to the framework
         // order's update list.  When the framework credit's
         // updateRecord() method is called, it will call the
         // cassette credit's updateRecord() method.
         Supervisor.getThreadCommitPoint().
            addToUpdateList(this.objFrameworkCredit);
      }
      this.objFrameworkCredit.setNewTransaction(false);

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.refund()"
         );
      }
   }

   /********************************************************************
   * This method is used by the order object when it is asked to service
   * a RefundReversal request.
   * <p>
   * For this cassette, the account object is used to void the
   * credit.  The framework credit object's state is updated and
   * return codes are placed in the provided response object.
   *
   *
   * @param     argRequest - A RefundReversal order request.
   * @param     argResponse - The response object to hold the result.
   * @exception ETillAbortOperation - any failure results in this
   * exception.
   ********************************************************************/
   public void refundReversal(
      RefundReversalRequest  argRequest,
      APIResponse            argResponse
   ) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.refundReversal()"
         );
      }

      // Abort RefundReversal if the Credit is not in refunded state.
      if (
          this.objFrameworkCredit.getCurrentState() != CREDIT_REFUNDED
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCredit RefundReversal was requested but the " +
               " credit is not in CREDIT_REFUNDED state." +
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_CREDIT +
               "\n\tState  = " + this.objFrameworkCredit.getCurrentState()
            );
         }
         throw new ETillAbortOperation(
            PRC_VERB_NOT_VALID_IN_PRESENT_STATE,
            RC_CREDIT
         );
      }
      // Obtain the framework's order object from the request.
      Order varFrameworkOrder = argRequest.getOrder();
      // Obtain the cassette order object from the framework order
      // object.
      LdbCardOrder varCassetteOrder =
         (LdbCardOrder)varFrameworkOrder.getCassetteOrder();
      // Obtain the account object that is used to process
      // transactions for this order.
      LdbCardAccount varAccount = varCassetteOrder.getAccount();
      // Ask the account to void the last credit.  If it returns
      // True, the void was approved.  Otherwise, the void was
      // was declined.  The account will set this object's approval
      // code and decline reason before it returns.
      if (varAccount.voidCredit(this)) {
         // Since the void was approved, remove this credit from
         // the current batch and add the framework batch to the
         // commit point.
         // Warning: the credit must be removed from the batch before
         // the framework credit object amount is set to zero.  In order
         // to update the batch credit totals correctly, the framework
         // batch subtracts the amount in the framework credit object.
         varAccount.retrieveBatchAndRemoveCredit(this.objFrameworkCredit);
         // Set the return codes in the response.
         // The void was approved.  Update the framework object,
         // add the credit to the batch, and set return codes.
         // Note:  The framework credit has a special integer
         // that contains the amount that has been refunded.  This
         // value is set to 0 here.
         this.objFrameworkCredit.setAmount(0);
         // Update the state information in the framework payment object.
         this.objFrameworkCredit.setCurrentState(CREDIT_VOID);
         argResponse.setReturnCodes(
            PRC_OPERATION_SUCCESS,
            RC_NONE
         );
      } else {
         // The void was declined.  The framework credit object
         // should remain in Refunded state, but the return codes
         // are set to indicate the failure.
         this.objFrameworkCredit.setCurrentState(CREDIT_REFUNDED);
         argResponse.setReturnCodes(
            PRC_FINANCIAL_FAILURE,
            RC_CASSETTE_DECLINED
         );
         // Trace the error.
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCredit Credit void declined. " +
               "\n\tPRC = " + PRC_FINANCIAL_FAILURE +
               "\n\tSRC = " + RC_CASSETTE_DECLINED +
               "\n\tReason = " + this.getReason()
            );
         }
      }
      // Add the framework credit object to the framework
      // order's update list.  When the framework credit's
      // updateRecord() method is called, it will call the
      // cassette credit's updateRecord() method.
      Supervisor.getThreadCommitPoint().
         addToUpdateList(this.objFrameworkCredit);

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.refundReversal()"
         );
      }
   }

   /********************************************************************
   * This method is used by the cassette order object when it is asked
   * to close a credit for a closing batch.
   ********************************************************************/
   public void closeCredit() {
      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionEntry(
            TRACE_ID,
            "LdbCardCredit.closeCredit()"
         );
      }

      // Assertion test.  If the credit is not in the correct state,
      // there is a serious defect and the cassette cannot continue.
      if (
         this.objFrameworkCredit.getCurrentState() != CREDIT_REFUNDED
      ) {
         if (Trace.isAnyoneTracing()) {
            Trace.traceErrorOccurred(
               TRACE_ID,
               "LdbCardCredit CloseCredit was requested but the" +
               " credit is not in CREDIT_REFUNDED state." +
               "\n\tPRC    = " + PRC_VERB_NOT_VALID_IN_PRESENT_STATE +
               "\n\tSRC    = " + RC_CREDIT +
               "\n\tState  = " + this.objFrameworkCredit.getCurrentState()
            );
         }
         Vector varVector = new Vector();
         varVector.addElement(
            this.objFrameworkCredit.getMerchantName()
         );
         varVector.addElement(
            this.objFrameworkCredit.getBatchNumber()
         );
         varVector.addElement(
            this.objFrameworkCredit.getOrderNumber()
         );
         varVector.addElement(
            this.objFrameworkCredit.getTransactionNumber()
         );
         varVector.addElement(
            Integer.toString(this.objFrameworkCredit.getCurrentState())
         );
         ErrorLog.log(
            RB_ID,
            MSG_CREDIT_CLOSE_STATE_ERROR,
            varVector
         );
         throw new ETillCassetteException(
            ErrorLog.lookupMessage(
               RB_ID,
               MSG_CREDIT_CLOSE_STATE_ERROR,
               varVector
            )
         );
      }

      // Set the credit state to closed and add the framework credit
      // to the update list.
      this.objFrameworkCredit.setCurrentState(CREDIT_CLOSED);
      Supervisor.getThreadCommitPoint().
         addToUpdateList(this.objFrameworkCredit);

      if (Trace.isAnyoneTracing()) {
         Trace.traceFunctionExit(
            TRACE_ID,
            "LdbCardCredit.closeCredit()"
         );
      }
   }

   /********************************************************************
   * This method is used to obtain the current approval code.
   *
   * @return    - The approval code returned by the payment engine.
   ********************************************************************/
   public String getApproval() {
      return this.objApprovalCode;
   }

   /********************************************************************
   * This method is used to obtain the framework's credit object.
   *
   * @return    - The framework's credit object for this credit.
   ********************************************************************/
   public Credit getFrameworkCredit() {
      return this.objFrameworkCredit;
   }

   /********************************************************************
   * This method is used to obtain the current reason code.
   *
   * @return    - The reason code (usually a text message) returned
   * by the payment engine when a transaction is declined.
   ********************************************************************/
   public String getReason() {
      return this.objDeclineReason;
   }

   /********************************************************************
   * This method is used to set the approval code for this credit.
   *
   * @param     argApprovalCode - The approval code returned by the
   * payment engine.
   ********************************************************************/
   public void setApproval(
      String argApprovalCode
   ) {
      this.objApprovalCode = argApprovalCode;
   }

   /********************************************************************
   * This method is used to set the decline reason for this credit.
   *
   * @param     argDeclineReason - The decline reason returned by the
   * payment engine.
   ********************************************************************/
   public void setReason(
      String argDeclineReason
   ) {
      this.objDeclineReason = argDeclineReason;
   }

}
